import argparse
import struct
import numpy as np
import sys
import os
from scipy.stats import spearmanr
from scipy.spatial.distance import cosine
import multiprocessing as mp


def parse_args(argv):
	parser = argparse.ArgumentParser(
		description='Transform binary vectors to text file.')

	parser.add_argument('input', type=str)
	parser.add_argument('--mode', type=str, default='r')
	parser.add_argument('--encoding', type=str, default='utf8')
	args = parser.parse_args(argv)

	return args


def read_vectors(fname, mode, encoding):
	word_vec = {}
	word_list = []
	with open(fname, mode, encoding=encoding) as fr:
		for i, line in enumerate(fr):
			if i == 0:
				a, b, c = map(int, line.split()[:3])
				print('Number of sememes: {}\n'
					  'Number of words: {}\n'
					  'Dimension of vectors: {}'.format(a, b, c))

			elif i <= b:
				word_list.append(line.split()[0])

			elif i <= a + b:
				continue

			else:
				parts = line.split()
				assert len(parts) % c == 0
				ix = 0
				for start in range(0, len(parts), c):
					cur_word = word_list[ix]
					word_vec[cur_word] = np.array(
						list(map(float, parts[start : start+c]))
					)

					if ix % 10000 == 0:
						print('{}K words read!'.format( int(ix/1000) ))
					ix += 1

	assert all([len(x)==c for x in word_vec.values()])
	return word_vec


def SimilarityTask(fname, vec):
	rank = []
	num = 0

	with open(fname, 'r', encoding='utf8') as fr:
		for i, line in enumerate(fr):
			w1, w2, _ = line.split()
			try:
				sim = cosine(vec[w1], vec[w2])
				rank.append((i, sim))
			except Exception:
				print('something is wrong with the {}-th line'.format(i+1))
			num += 1

	rank_sorted = sorted(rank, key=lambda x: x[1])
	rank_pos = [x[0] for x in rank_sorted]
	try:
		corr, _ = spearmanr(rank_pos, range(num))
	except Exception:
		corr, _ = spearmanr(rank_pos, sorted(rank_pos))
	finally:
		print('The correlation is {} for task {}'.format(
			corr, fname)
		)


class Eval:
	def __init__(self, ctg):
		self.ctg = ctg
		self.acc = 0
		self.rank = 0
		self.num = 0
	def update(self, rank):
		if rank == 1:
			self.acc += 1
		self.rank += rank
		self.num += 1
	def evaluate(self):
		acc = self.acc / self.num
		mean_rank = self.rank / self.num
		print('Category: {}\n'
			  'Total count: {}\n'
			  'Accuracy: {}\n'
			  'Mean rank: {}\n'.format(self.ctg, self.num, acc, mean_rank))


def worker(fname, target, vec, queue):
	line_no = 0
	result = Eval(target)

	with open(fname, 'r', encoding='utf8') as fr:
		
		while True:
			line = fr.readline()
			line_no += 1
			
			if not line:
				print('target {} not found'.format(target))
				break
			
			if line[0] != ':':
				continue
			
			topic = line.split()[1].split('-')[0]
			if topic == target:
				print('target {} found. Beginning...'.format(target))
				break

		while True:
			line = fr.readline()
			line_no += 1

			if not line or line[0] == ':':
				print('target {} finished'.format(target))
				break

			words = line.split()
			assert len(words) == 4

			if any([w not in vec for w in words]):
				print('something is wrong with the {}-th line'.format(line_no))
				continue

			v1 = vec[words[1]] - vec[words[0]] + vec[words[2]]
			v2 = vec[words[-1]]

			v2_rank = 1
			v2_score = cosine(v1, v2)

			for w, v in vec.items():
				if w in words:
					continue
				score = cosine(v1, v)
				if score < v2_score:
					v2_rank += 1

			result.update(v2_rank)

	queue.put(result)


def AnalogyTask(fname, vec):
	queue = mp.Queue()
	category = ['capital', 'city', 'family']
	processes = [mp.Process(target=worker, args=(fname, x, vec, queue)) for x in category]

	for p in processes:
		p.start()

	for p in processes:
		p.join()

	results = [queue.get() for p in processes]
	for res in results:
		res.evaluate()

	acc, rank, num = 0, 0, 0
	for res in results:
		acc += res.acc
		rank += res.rank
		num += res.num
	print('Total acc: {}\nTotal mean rank: {}\n'
		  'Total number: {}\n'.format(acc / num,
		  							  rank / num,
		  							  num))
	return results


def main(argv):
	args = parse_args(argv)

	assert os.path.isfile(args.input)

	word_vec = read_vectors(args.input, args.mode, args.encoding)

	task_folder = r'path/to/the/folder/where/tasks/are/placed'
	SimilarityTask(os.path.join(task_folder, 'wordsim-240.txt'), word_vec)
	SimilarityTask(os.path.join(task_folder, 'wordsim-297.txt'), word_vec)
	results = AnalogyTask(os.path.join(task_folder, 'analogy.txt'), word_vec)


if __name__ == '__main__':
	main(sys.argv[1:])
